#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "adaptivetimestepper.hh"

void IterationRegister::registerCount(FixedPointIterationCounter count) {
  totalCount += count;
}

void IterationRegister::registerFinalCount(FixedPointIterationCounter count) {
  finalCount = count;
}

void IterationRegister::reset() {
  totalCount = FixedPointIterationCounter();
  finalCount = FixedPointIterationCounter();
}

template <class Factory, class NBodyAssembler, class Updaters, class ErrorNorm>
AdaptiveTimeStepper<Factory, NBodyAssembler, Updaters, ErrorNorm>::AdaptiveTimeStepper(
        Dune::ParameterTree const &parset,
        NBodyAssembler& nBodyAssembler,
        const IgnoreVector& ignoreNodes,
        GlobalFriction& globalFriction,
        const std::vector<const BitVector*>& bodywiseNonmortarBoundaries,
        Updaters &current,
        double relativeTime,
        double relativeTau,
        ExternalForces& externalForces,
        const std::vector<const ErrorNorm* >& errorNorms,
        std::function<bool(Updaters &, Updaters &)> mustRefine)
    : relativeTime_(relativeTime),
      relativeTau_(relativeTau),
      finalTime_(parset.get<double>("problem.finalTime")),
      parset_(parset),
      nBodyAssembler_(nBodyAssembler),
      ignoreNodes_(ignoreNodes),
      globalFriction_(globalFriction),
      bodywiseNonmortarBoundaries_(bodywiseNonmortarBoundaries),
      current_(current),
      R1_(),
      externalForces_(externalForces),
      mustRefine_(mustRefine),
      errorNorms_(errorNorms) {}

template <class Factory, class NBodyAssembler, class Updaters, class ErrorNorm>
bool AdaptiveTimeStepper<Factory, NBodyAssembler, Updaters, ErrorNorm>::reachedEnd() {
  return relativeTime_ >= 1.0;
}

template <class Factory, class NBodyAssembler, class Updaters, class ErrorNorm>
IterationRegister AdaptiveTimeStepper<Factory, NBodyAssembler, Updaters, ErrorNorm>::advance() {
  /*
    |     C     | We check here if making the step R1 of size tau is a
    |  R1 | R2  | good idea. To check if we can coarsen, we compare
    |F1|F2|     | the result of (R1+R2) with C, i.e. two steps of size
                  tau with one of size 2*tau. To check if we need to
    refine, we compare the result of (F1+F2) with R1, i.e. two steps
    of size tau/2 with one of size tau. The method makes multiple
    coarsening/refining attempts, with coarsening coming first. */

  std::cout << "AdaptiveTimeStepper::advance()" << std::endl;

  if (R1_.updaters == Updaters())
    R1_ = step(current_, relativeTime_, relativeTau_);

  std::cout << "AdaptiveTimeStepper Step 1" << std::endl;


  bool didCoarsen = false;
  iterationRegister_.reset();
  UpdatersWithCount R2;
  UpdatersWithCount C;
  while (relativeTime_ + relativeTau_ <= 1.0) {
    R2 = step(R1_.updaters, relativeTime_ + relativeTau_, relativeTau_);
    std::cout << "AdaptiveTimeStepper R2 computed!" << std::endl << std::endl;
    C = step(current_, relativeTime_, 2 * relativeTau_);
    std::cout << "AdaptiveTimeStepper C computed!" << std::endl << std::endl;
    if (mustRefine_(R2.updaters, C.updaters))
      break;

    didCoarsen = true;
    relativeTau_ *= 2;
    R1_ = C;
  }

  std::cout << "AdaptiveTimeStepper Step 1" << std::endl;
  UpdatersWithCount F1;
  UpdatersWithCount F2;
  if (!didCoarsen) {
    while (true) {
      F1 = step(current_, relativeTime_, relativeTau_ / 2.0);
      std::cout << "AdaptiveTimeStepper F1 computed!" << std::endl << std::endl;
      F2 = step(F1.updaters, relativeTime_ + relativeTau_ / 2.0,
                relativeTau_ / 2.0);
      std::cout << "AdaptiveTimeStepper F2 computed!" << std::endl << std::endl;
      if (!mustRefine_(F2.updaters, R1_.updaters)) {
        std::cout << "Sufficiently refined!" << std::endl;
        break;
      }

      relativeTau_ /= 2.0;
      R1_ = F1;
      R2 = F2;
    }
  }

  std::cout << "AdaptiveTimeStepper::advance() ...";

  iterationRegister_.registerFinalCount(R1_.count);
  relativeTime_ += relativeTau_;
  current_ = R1_.updaters;

  //UpdatersWithCount emptyR1;
  //R1_ = emptyR1;
  R1_ = R2;

  std::cout << " done" << std::endl;

  return iterationRegister_;
}

template <class Factory, class NBodyAssembler, class Updaters, class ErrorNorm>
typename AdaptiveTimeStepper<Factory, NBodyAssembler, Updaters, ErrorNorm>::UpdatersWithCount
AdaptiveTimeStepper<Factory, NBodyAssembler, Updaters, ErrorNorm>::step(
    Updaters const &oldUpdaters, double rTime, double rTau) {
  UpdatersWithCount newUpdatersAndCount = {oldUpdaters.clone(), {}};
  newUpdatersAndCount.count =
      MyCoupledTimeStepper(finalTime_, parset_, nBodyAssembler_, ignoreNodes_, globalFriction_, bodywiseNonmortarBoundaries_,
                           newUpdatersAndCount.updaters, errorNorms_,
                           externalForces_)
          .step(rTime, rTau);
  iterationRegister_.registerCount(newUpdatersAndCount.count);
  return newUpdatersAndCount;
}

#include "adaptivetimestepper_tmpl.cc"
